@@ -1,5 +1,5 @@
-// Copyright (C) 2011-2013 Conrad Sanderson
-// Copyright (C) 2011-2013 NICTA (www.nicta.com.au)
+// Copyright (C) 2011-2015 Conrad Sanderson
+// Copyright (C) 2011-2015 NICTA (www.nicta.com.au)
 // 
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -17,7 +17,7 @@ arma_inline
 void
 arrayops::copy(eT* dest, const eT* src, const uword n_elem)
   {
-  if(n_elem <= 16)
+  if( (n_elem <= 16) && (is_cx<eT>::no) )
     {
     arrayops::copy_small(dest, src, n_elem);
     }
@@ -66,44 +66,25 @@ void
 arrayops::copy_forwards(eT* dest, const eT* src, const uword n_elem)
   {
   // can't use std::memcpy(), as we don't know how it copies data
-  uword i,j;
+  uword j;
   
-  for(i=0, j=1; j < n_elem; i+=2, j+=2)
+  for(j=1; j < n_elem; j+=2)
     {
-    dest[i] = src[i];
-    dest[j] = src[j];
+    const eT tmp_i = (*src);  src++;
+    const eT tmp_j = (*src);  src++;
+    
+    (*dest) = tmp_i;  dest++;
+    (*dest) = tmp_j;  dest++;
     }
   
-  if(i < n_elem)
+  if((j-1) < n_elem)
     {
-    dest[i] = src[i];
+    (*dest) = (*src);
     }
   }
 
 
 
-// template<typename eT>
-// arma_hot
-// inline
-// void
-// arrayops::copy_backwards(eT* dest, const eT* src, const uword n_elem)
-//   {
-//   // can't use std::memcpy(), as we don't know how it copies data
-//   
-//   switch(n_elem)
-//     {
-//     default:
-//       for(uword i = (n_elem-1); i >= 1; --i) { dest[i] = src[i]; }
-//       // NOTE: the 'break' statement has been deliberately omitted
-//     case 1:
-//       dest[0] = src[0];
-//     case 0:
-//       ;
-//     }
-//   }
-
-
-
 template<typename eT>
 arma_hot
 inline
@@ -112,35 +93,32 @@ arrayops::copy_backwards(eT* dest, const eT* src, const uword n_elem)
   {
   // can't use std::memcpy(), as we don't know how it copies data
   
-  switch(n_elem)
+  // for(uword i=0; i < n_elem; ++i) 
+  //   {
+  //   const uword j = n_elem-i-1;
+  //   
+  //   dest[j] = src[j];
+  //   }
+  
+  if(n_elem > 0)
     {
-    default:
+          eT* dest_it = &(dest[n_elem-1]);
+    const eT*  src_it = &( src[n_elem-1]);
+    
+    uword j;
+    for(j=1; j < n_elem; j+=2) 
       {
-      uword i, j;
-      for(i = (n_elem-1), j = (n_elem-2); j >= 2; i-=2, j-=2)
-        {
-        const eT tmp_i = src[i];
-        const eT tmp_j = src[j];
-        
-        dest[i] = tmp_i;
-        dest[j] = tmp_j;
-        }
+      const eT tmp_i = (*src_it);  src_it--;
+      const eT tmp_j = (*src_it);  src_it--;
       
-      // j is less than 2:  it can be 1 or 0
-      // i is j+1, ie. less than 3: it can be 2 or 1
-      
-      if(i == 2)
-        {
-        dest[2] = src[2];
-        }
+      (*dest_it) = tmp_i;  dest_it--;
+      (*dest_it) = tmp_j;  dest_it--;
+      }
+    
+    if((j-1) < n_elem)
+      {
+      (*dest_it) = (*src_it);
       }
-      // NOTE: the 'break' statement has been deliberately omitted
-    case 2:
-      dest[1] = src[1];
-    case 1:
-      dest[0] = src[0];
-    case 0:
-      ;
     }
   }
 
@@ -152,16 +130,7 @@ inline
 void
 arrayops::fill_zeros(eT* dest, const uword n_elem)
   {
-  typedef typename get_pod_type<eT>::result pod_type;
-  
-  if( (n_elem >= 8) && (std::numeric_limits<eT>::is_integer || (std::numeric_limits<pod_type>::is_iec559 && is_real<pod_type>::value)) )
-    {
-    std::memset(dest, 0, sizeof(eT)*n_elem);
-    }
-  else
-    {
-    arrayops::inplace_set(dest, eT(0), n_elem);
-    }
+  arrayops::inplace_set(dest, eT(0), n_elem);
   }
 
 
@@ -227,32 +196,45 @@ inline
 void
 arrayops::convert(out_eT* dest, const in_eT* src, const uword n_elem)
   {
-  uword i,j;
-  
-  for(i=0, j=1; j<n_elem; i+=2, j+=2)
+  if(is_same_type<out_eT,in_eT>::value)
     {
-    const in_eT tmp_i = src[i];
-    const in_eT tmp_j = src[j];
+    const out_eT* src2 = (const out_eT*)src;
+    
+    if(dest != src2)  { arrayops::copy(dest, src2, n_elem); }
+    
+    return;
+    }
+  
+  
+  uword j;
+  
+  for(j=1; j<n_elem; j+=2)
+    {
+    const in_eT tmp_i = (*src);  src++;
+    const in_eT tmp_j = (*src);  src++;
     
     // dest[i] = out_eT( tmp_i );
     // dest[j] = out_eT( tmp_j );
     
-    dest[i] = (is_signed<out_eT>::value)
+    (*dest) = (is_signed<out_eT>::value)
               ? out_eT( tmp_i )
               : ( cond_rel< is_signed<in_eT>::value >::lt(tmp_i, in_eT(0)) ? out_eT(0) : out_eT(tmp_i) );
-              
-    dest[j] = (is_signed<out_eT>::value)
+    
+    dest++;
+    
+    (*dest) = (is_signed<out_eT>::value)
               ? out_eT( tmp_j )
               : ( cond_rel< is_signed<in_eT>::value >::lt(tmp_j, in_eT(0)) ? out_eT(0) : out_eT(tmp_j) );
+    dest++;
     }
   
-  if(i < n_elem)
+  if((j-1) < n_elem)
     {
-    const in_eT tmp_i = src[i];
+    const in_eT tmp_i = (*src);
     
     // dest[i] = out_eT( tmp_i );
     
-    dest[i] = (is_signed<out_eT>::value)
+    (*dest) = (is_signed<out_eT>::value)
               ? out_eT( tmp_i )
               : ( cond_rel< is_signed<in_eT>::value >::lt(tmp_i, in_eT(0)) ? out_eT(0) : out_eT(tmp_i) );
     }
@@ -266,17 +248,17 @@ inline
 void
 arrayops::convert_cx(out_eT* dest, const in_eT* src, const uword n_elem)
   {
-  uword i,j;
+  uword j;
   
-  for(i=0, j=1; j<n_elem; i+=2, j+=2)
+  for(j=1; j<n_elem; j+=2)
     {
-    arrayops::convert_cx_scalar( dest[i], src[i] );
-    arrayops::convert_cx_scalar( dest[j], src[j] );
+    arrayops::convert_cx_scalar( (*dest), (*src) );  dest++; src++;
+    arrayops::convert_cx_scalar( (*dest), (*src) );  dest++; src++;
     }
   
-  if(i < n_elem)
+  if((j-1) < n_elem)
     {
-    arrayops::convert_cx_scalar( dest[i], src[i] );
+    arrayops::convert_cx_scalar( (*dest), (*src) );
     }
   }
 
@@ -440,21 +422,32 @@ inline
 void
 arrayops::inplace_plus_base(eT* dest, const eT* src, const uword n_elem)
   {
-  uword i,j;
-  
-  for(i=0, j=1; j<n_elem; i+=2, j+=2)
+  #if defined(ARMA_SIMPLE_LOOPS)
     {
-    const eT tmp_i = src[i];
-    const eT tmp_j = src[j];
+    for(uword i=0; i<n_elem; ++i)
+      {
+      dest[i] += src[i];
+      }
+    }
+  #else
+    {
+    uword i,j;
     
-    dest[i] += tmp_i;
-    dest[j] += tmp_j;
-    }
-  
-  if(i < n_elem)
-    {
-    dest[i] += src[i];
+    for(i=0, j=1; j<n_elem; i+=2, j+=2)
+      {
+      const eT tmp_i = src[i];
+      const eT tmp_j = src[j];
+      
+      dest[i] += tmp_i;
+      dest[j] += tmp_j;
+      }
+    
+    if(i < n_elem)
+      {
+      dest[i] += src[i];
+      }
     }
+  #endif
   }
 
 
@@ -465,21 +458,32 @@ inline
 void
 arrayops::inplace_minus_base(eT* dest, const eT* src, const uword n_elem)
   {
-  uword i,j;
-  
-  for(i=0, j=1; j<n_elem; i+=2, j+=2)
+  #if defined(ARMA_SIMPLE_LOOPS)
     {
-    const eT tmp_i = src[i];
-    const eT tmp_j = src[j];
+    for(uword i=0; i<n_elem; ++i)
+      {
+      dest[i] -= src[i];
+      }
+    }
+  #else
+    {
+    uword i,j;
     
-    dest[i] -= tmp_i;
-    dest[j] -= tmp_j;
-    }
-  
-  if(i < n_elem)
-    {
-    dest[i] -= src[i];
+    for(i=0, j=1; j<n_elem; i+=2, j+=2)
+      {
+      const eT tmp_i = src[i];
+      const eT tmp_j = src[j];
+      
+      dest[i] -= tmp_i;
+      dest[j] -= tmp_j;
+      }
+    
+    if(i < n_elem)
+      {
+      dest[i] -= src[i];
+      }
     }
+  #endif
   }
 
 
@@ -490,21 +494,32 @@ inline
 void
 arrayops::inplace_mul_base(eT* dest, const eT* src, const uword n_elem)
   {
-  uword i,j;
-  
-  for(i=0, j=1; j<n_elem; i+=2, j+=2)
+  #if defined(ARMA_SIMPLE_LOOPS)
     {
-    const eT tmp_i = src[i];
-    const eT tmp_j = src[j];
+    for(uword i=0; i<n_elem; ++i)
+      {
+      dest[i] *= src[i];
+      }
+    }
+  #else
+    {
+    uword i,j;
     
-    dest[i] *= tmp_i;
-    dest[j] *= tmp_j;
-    }
-  
-  if(i < n_elem)
-    {
-    dest[i] *= src[i];
+    for(i=0, j=1; j<n_elem; i+=2, j+=2)
+      {
+      const eT tmp_i = src[i];
+      const eT tmp_j = src[j];
+      
+      dest[i] *= tmp_i;
+      dest[j] *= tmp_j;
+      }
+    
+    if(i < n_elem)
+      {
+      dest[i] *= src[i];
+      }
     }
+  #endif
   }
 
 
@@ -515,21 +530,32 @@ inline
 void
 arrayops::inplace_div_base(eT* dest, const eT* src, const uword n_elem)
   {
-  uword i,j;
-  
-  for(i=0, j=1; j<n_elem; i+=2, j+=2)
+  #if defined(ARMA_SIMPLE_LOOPS)
     {
-    const eT tmp_i = src[i];
-    const eT tmp_j = src[j];
+    for(uword i=0; i<n_elem; ++i)
+      {
+      dest[i] /= src[i];
+      }
+    }
+  #else
+    {
+    uword i,j;
     
-    dest[i] /= tmp_i;
-    dest[j] /= tmp_j;
-    }
-  
-  if(i < n_elem)
-    {
-    dest[i] /= src[i];
+    for(i=0, j=1; j<n_elem; i+=2, j+=2)
+      {
+      const eT tmp_i = src[i];
+      const eT tmp_j = src[j];
+      
+      dest[i] /= tmp_i;
+      dest[j] /= tmp_j;
+      }
+    
+    if(i < n_elem)
+      {
+      dest[i] /= src[i];
+      }
     }
+  #endif
   }
 
 
@@ -540,10 +566,51 @@ inline
 void
 arrayops::inplace_set(eT* dest, const eT val, const uword n_elem)
   {
-  if(memory::is_aligned(dest))
+  typedef typename get_pod_type<eT>::result pod_type;
+  
+  if( (n_elem <= 16) && (is_cx<eT>::no) )
+    {
+    arrayops::inplace_set_small(dest, val, n_elem);
+    }
+  else
+    {
+    if( (val == eT(0)) && (std::numeric_limits<eT>::is_integer || (std::numeric_limits<pod_type>::is_iec559 && is_real<pod_type>::value)) )
+      {
+      std::memset(dest, 0, sizeof(eT)*n_elem);
+      }
+    else
+      {
+      if(memory::is_aligned(dest))
+        {
+        memory::mark_as_aligned(dest);
+        
+        arrayops::inplace_set_base(dest, val, n_elem);
+        }
+      else
+        {
+        arrayops::inplace_set_base(dest, val, n_elem);
+        }
+      }
+    }
+  }
+
+
+
+template<typename eT>
+arma_hot
+inline
+void
+arrayops::inplace_set_base(eT* dest, const eT val, const uword n_elem)
+  {
+  #if defined(ARMA_SIMPLE_LOOPS)
+    {
+    for(uword i=0; i<n_elem; ++i)
+      {
+      dest[i] = val;
+      }
+    }
+  #else
     {
-    memory::mark_as_aligned(dest);
-    
     uword i,j;
     
     for(i=0, j=1; j<n_elem; i+=2, j+=2)
@@ -557,21 +624,37 @@ arrayops::inplace_set(eT* dest, const eT val, const uword n_elem)
       dest[i] = val;
       }
     }
-  else
+  #endif
+  }
+
+
+
+template<typename eT>
+arma_hot
+inline
+void
+arrayops::inplace_set_small(eT* dest, const eT val, const uword n_elem)
+  {
+  switch(n_elem)
     {
-    uword i,j;
-    
-    for(i=0, j=1; j<n_elem; i+=2, j+=2)
-      {
-      dest[i] = val;
-      dest[j] = val;
-      }
-    
-    if(i < n_elem)
-      {
-      dest[i] = val;
-      }
-    } 
+    case 16: dest[15] = val;
+    case 15: dest[14] = val;
+    case 14: dest[13] = val;
+    case 13: dest[12] = val;
+    case 12: dest[11] = val;
+    case 11: dest[10] = val;
+    case 10: dest[ 9] = val;
+    case  9: dest[ 8] = val;
+    case  8: dest[ 7] = val;
+    case  7: dest[ 6] = val;
+    case  6: dest[ 5] = val;
+    case  5: dest[ 4] = val;
+    case  4: dest[ 3] = val;
+    case  3: dest[ 2] = val;
+    case  2: dest[ 1] = val;
+    case  1: dest[ 0] = val;
+    default:;
+    }
   }
 
 
@@ -600,33 +683,11 @@ arrayops::inplace_plus(eT* dest, const eT val, const uword n_elem)
     {
     memory::mark_as_aligned(dest);
     
-    uword i,j;
-    
-    for(i=0, j=1; j<n_elem; i+=2, j+=2)
-      {
-      dest[i] += val;
-      dest[j] += val;
-      }
-    
-    if(i < n_elem)
-      {
-      dest[i] += val;
-      }
+    arrayops::inplace_plus_base(dest, val, n_elem);
     }
   else
     {
-    uword i,j;
-    
-    for(i=0, j=1; j<n_elem; i+=2, j+=2)
-      {
-      dest[i] += val;
-      dest[j] += val;
-      }
-    
-    if(i < n_elem)
-      {
-      dest[i] += val;
-      }
+    arrayops::inplace_plus_base(dest, val, n_elem);
     }
   }
 
@@ -642,33 +703,11 @@ arrayops::inplace_minus(eT* dest, const eT val, const uword n_elem)
     {
     memory::mark_as_aligned(dest);
     
-    uword i,j;
-    
-    for(i=0, j=1; j<n_elem; i+=2, j+=2)
-      {
-      dest[i] -= val;
-      dest[j] -= val;
-      }
-    
-    if(i < n_elem)
-      {
-      dest[i] -= val;
-      }
+    arrayops::inplace_minus_base(dest, val, n_elem);
     }
   else
     {
-    uword i,j;
-    
-    for(i=0, j=1; j<n_elem; i+=2, j+=2)
-      {
-      dest[i] -= val;
-      dest[j] -= val;
-      }
-    
-    if(i < n_elem)
-      {
-      dest[i] -= val;
-      }
+    arrayops::inplace_minus_base(dest, val, n_elem);
     }
   }
 
@@ -684,33 +723,11 @@ arrayops::inplace_mul(eT* dest, const eT val, const uword n_elem)
     {
     memory::mark_as_aligned(dest);
     
-    uword i,j;
-    
-    for(i=0, j=1; j<n_elem; i+=2, j+=2)
-      {
-      dest[i] *= val;
-      dest[j] *= val;
-      }
-    
-    if(i < n_elem)
-      {
-      dest[i] *= val;
-      }
+    arrayops::inplace_mul_base(dest, val, n_elem);
     }
   else
     {
-    uword i,j;
-    
-    for(i=0, j=1; j<n_elem; i+=2, j+=2)
-      {
-      dest[i] *= val;
-      dest[j] *= val;
-      }
-    
-    if(i < n_elem)
-      {
-      dest[i] *= val;
-      }
+    arrayops::inplace_mul_base(dest, val, n_elem);
     }
   }
 
@@ -726,20 +743,129 @@ arrayops::inplace_div(eT* dest, const eT val, const uword n_elem)
     {
     memory::mark_as_aligned(dest);
     
+    arrayops::inplace_div_base(dest, val, n_elem);
+    }
+  else
+    {
+    arrayops::inplace_div_base(dest, val, n_elem);
+    }
+  }
+
+
+
+template<typename eT>
+arma_hot
+inline
+void
+arrayops::inplace_plus_base(eT* dest, const eT val, const uword n_elem)
+  {
+  #if defined(ARMA_SIMPLE_LOOPS)
+    {
+    for(uword i=0; i<n_elem; ++i)
+      {
+      dest[i] += val;
+      }
+    }
+  #else
+    {
     uword i,j;
     
     for(i=0, j=1; j<n_elem; i+=2, j+=2)
       {
-      dest[i] /= val;
-      dest[j] /= val;
+      dest[i] += val;
+      dest[j] += val;
       }
     
     if(i < n_elem)
+      {
+      dest[i] += val;
+      }
+    }
+  #endif
+  }
+
+
+
+template<typename eT>
+arma_hot
+inline
+void
+arrayops::inplace_minus_base(eT* dest, const eT val, const uword n_elem)
+  {
+  #if defined(ARMA_SIMPLE_LOOPS)
+    {
+    for(uword i=0; i<n_elem; ++i)
+      {
+      dest[i] -= val;
+      }
+    }
+  #else
+    {
+    uword i,j;
+    
+    for(i=0, j=1; j<n_elem; i+=2, j+=2)
+      {
+      dest[i] -= val;
+      dest[j] -= val;
+      }
+    
+    if(i < n_elem)
+      {
+      dest[i] -= val;
+      }
+    }
+  #endif
+  }
+
+
+
+template<typename eT>
+arma_hot
+inline
+void
+arrayops::inplace_mul_base(eT* dest, const eT val, const uword n_elem)
+  {
+  #if defined(ARMA_SIMPLE_LOOPS)
+    {
+    for(uword i=0; i<n_elem; ++i)
+      {
+      dest[i] *= val;
+      }
+    }
+  #else
+    {
+    uword i,j;
+    
+    for(i=0, j=1; j<n_elem; i+=2, j+=2)
+      {
+      dest[i] *= val;
+      dest[j] *= val;
+      }
+    
+    if(i < n_elem)
+      {
+      dest[i] *= val;
+      }
+    }
+  #endif
+  }
+
+
+
+template<typename eT>
+arma_hot
+inline
+void
+arrayops::inplace_div_base(eT* dest, const eT val, const uword n_elem)
+  {
+  #if defined(ARMA_SIMPLE_LOOPS)
+    {
+    for(uword i=0; i<n_elem; ++i)
       {
       dest[i] /= val;
       }
     }
-  else
+  #else
     {
     uword i,j;
     
@@ -754,6 +880,7 @@ arrayops::inplace_div(eT* dest, const eT val, const uword n_elem)
       dest[i] /= val;
       }
     }
+  #endif
   }
 
 
@@ -765,23 +892,43 @@ inline
 eT
 arrayops::accumulate(const eT* src, const uword n_elem)
   {
-  uword i,j;
-  
-  eT acc1 = eT(0);
-  eT acc2 = eT(0);
-  
-  for(i=0, j=1; j<n_elem; i+=2, j+=2)
+  #if defined(__FINITE_MATH_ONLY__) && (__FINITE_MATH_ONLY__ > 0)
     {
-    acc1 += src[i];
-    acc2 += src[j];
+    eT acc = eT(0);
+    
+    if(memory::is_aligned(src))
+      {
+      memory::mark_as_aligned(src);
+      for(uword i=0; i<n_elem; ++i)  { acc += src[i]; }
+      }
+    else
+      {
+      for(uword i=0; i<n_elem; ++i)  { acc += src[i]; }
+      }
+    
+    return acc;
     }
-  
-  if(i < n_elem)
+  #else
     {
-    acc1 += src[i];
+    eT acc1 = eT(0);
+    eT acc2 = eT(0);
+    
+    uword j;
+    
+    for(j=1; j<n_elem; j+=2)
+      {
+      acc1 += (*src);  src++;
+      acc2 += (*src);  src++;
+      }
+    
+    if((j-1) < n_elem)
+      {
+      acc1 += (*src);
+      }
+    
+    return acc1 + acc2;
     }
-  
-  return acc1 + acc2;
+  #endif
   }
 
 
@@ -821,12 +968,12 @@ inline
 bool
 arrayops::is_finite(const eT* src, const uword n_elem)
   {
-  uword i,j;
+  uword j;
   
-  for(i=0, j=1; j<n_elem; i+=2, j+=2)
+  for(j=1; j<n_elem; j+=2)
     {
-    const eT val_i = src[i];
-    const eT val_j = src[j];
+    const eT val_i = (*src);  src++;
+    const eT val_j = (*src);  src++;
     
     if( (arma_isfinite(val_i) == false) || (arma_isfinite(val_j) == false) )
       {
@@ -834,9 +981,9 @@ arrayops::is_finite(const eT* src, const uword n_elem)
       }
     }
   
-  if(i < n_elem)
+  if((j-1) < n_elem)
     {
-    if(arma_isfinite(src[i]) == false)
+    if(arma_isfinite(*src) == false)
       {
       return false;
       }
